# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""UI layer for invenio-workflows.

workflows UI is a web interface overlay for all WorkflowObject's.

This area is targeted to catalogers and administrators for inspecting
and reacting to workflows executions. More importantly, allowing users to deal
with halted workflows.

For example, accepting submissions or other tasks.
"""

from __future__ import absolute_import, print_function

from flask import (
    Blueprint,
    abort,
    render_template,
    current_app
)

from flask_login import login_required

from invenio_workflows.errors import WorkflowsMissingObject

from invenio_workflows.proxies import workflow_object_class

from ..utils import (
    obj_or_import_string
)
from ..permissions import admin_permission_factory


def create_blueprint(config, url_endpoint, context_processors):
    """Create UI blueprint for invenio-workflows-ui."""
    blueprint = Blueprint(
        'invenio_workflows_ui',
        __name__,
        url_prefix=url_endpoint,
        template_folder='../templates',
        static_folder='../static',
    )

    @blueprint.route('/', methods=['GET', 'POST'])
    @blueprint.route('/index', methods=['GET', 'POST'])
    @login_required
    def index():
        """Display basic dashboard interface of Workflows UI."""
        if not admin_permission_factory().can():
            abort(403)
        return render_template(
            current_app.config['WORKFLOWS_UI_INDEX_TEMPLATE']
        )

    @blueprint.route('/list', methods=['GET', ])
    @blueprint.route('/list/', methods=['GET', ])
    @blueprint.route('/list/<search_value>', methods=['GET', ])
    @login_required
    def list_objects(search_value=None):
        """Display main table interface of workflows UI."""
        if not admin_permission_factory().can():
            abort(403)
        return render_template(
            current_app.config['WORKFLOWS_UI_LIST_TEMPLATE'],
            search=search_value
        )

    @blueprint.route('/<int:objectid>', methods=['GET', 'POST'])
    @blueprint.route('/details/<int:objectid>', methods=['GET', 'POST'])
    @login_required
    def details(objectid):
        """Display info about the object."""
        if not admin_permission_factory().can():
            abort(403)
        try:
            workflow_object = workflow_object_class.get(objectid)
        except WorkflowsMissingObject:
            abort(404)

        return render_template(
            current_app.config['WORKFLOWS_UI_DETAILS_TEMPLATE'],
            workflow_object=workflow_object,
        )

    for proc in context_processors:
        blueprint.context_processor(obj_or_import_string(proc))

    return blueprint



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""API."""

from __future__ import absolute_import, print_function

from functools import partial, wraps

import six
from elasticsearch import TransportError
from flask import current_app, request
from invenio_db import db
from invenio_records import Record
from invenio_records.errors import MissingModelError
from invenio_workflows import ObjectStatus, resume
from invenio_workflows.proxies import workflow_object_class, workflows

from .indexer import WorkflowIndexer
from .proxies import actions


def record_to_index(record):
    """Index the workflow record into desired index/doc_type."""
    config = current_app.config['WORKFLOWS_UI_DATA_TYPES'].get(
        record["_workflow"]["data_type"]
    )
    if not config:
        return None, None
    return config.get('search_index'), config.get('search_type')


def index(method=None, delete=False):
    """Apply to API methods that need to change the index for the object."""
    # Check if we shall save arguments and recreate decorator
    if method is None:
        return partial(index, delete=delete)

    @wraps(method)
    def wrapper(self_or_cls, *args, **kwargs):
        """Send record for indexing."""
        result = method(self_or_cls, *args, **kwargs)
        try:
            if delete:
                self_or_cls.indexer.delete(result)
            else:
                self_or_cls.indexer.index(result)
        except TransportError as err:
            current_app.logger.exception(err)
            current_app.logger.error(
                "Problem while indexing workflow object {0}".format(
                    self_or_cls.id
                )
            )
        return result
    return wrapper


class WorkflowUIRecord(Record):
    """Represents a workflow object record for indexing."""

    indexer = WorkflowIndexer(
        record_to_index=record_to_index,
    )

    def __init__(self, *args, **kwargs):
        """Represent a workflow object record for indexing."""
        try:
            self.workflow = kwargs.pop('workflow')
        except KeyError:
            raise TypeError(
                'WorkflowUIRecord.__init__ missing workflow argument'
            )

        kwargs['model'] = self.workflow.model
        super(WorkflowUIRecord, self).__init__(*args, **kwargs)

    @classmethod
    @index
    def create(cls, workflow_object, **kwargs):
        """Create a indexable workflow JSON."""
        record = cls.record_from_object(workflow_object)
        return cls(record, workflow=workflow_object, **kwargs)

    @classmethod
    def get_record(cls, id_, with_deleted=False):
        """Get record instance.

        Raises database exception if record does not exists.
        """
        with db.session.no_autoflush:
            obj = workflow_object_class.get(id_)
            return cls(cls.record_from_object(obj), workflow=obj)

    def commit(self):
        """Commit a change to the record state."""
        with db.session.begin_nested():
            self.update_model()

    def delete(self, force=False):
        """Delete model from DB and search index.

        The index is deleted using a signal, see the `receivers` module).
        """
        if self.model is None:
            raise MissingModelError()

        with db.session.begin_nested():
            db.session.delete(self.model)
        return self

    @staticmethod
    def record_from_object(workflow_object):
        """Build data from workflow object.

        NOTE: This entire function may in principle be in
        workflow_object_class model as a to_dict() kind of function of
        the model.
        """
        record = {}
        record["id"] = workflow_object.id
        _workflow = {}
        _workflow["data_type"] = workflow_object.data_type
        _workflow["status"] = workflow_object.status.name
        _workflow["id_user"] = workflow_object.id_user
        _workflow["id_parent"] = workflow_object.id_parent
        _workflow["id_workflow"] = None
        _workflow["workflow_class"] = None
        _workflow["workflow_position"] = workflow_object.callback_pos
        _workflow["workflow_name"] = None

        if (
                workflow_object.workflow and
                workflow_object.workflow.name in workflows
        ):
            workflow_definition = workflows.get(workflow_object.workflow.name)

            if (
                    not _workflow["data_type"] and
                    workflow_definition and
                    hasattr(workflow_definition, 'data_type')
            ):
                _workflow["data_type"] = workflow_definition.data_type

            if workflow_definition and hasattr(workflow_definition, 'name'):
                _workflow["workflow_name"] = workflow_definition.name

            if workflow_object.id_workflow:
                _workflow["id_workflow"] = six.text_type(
                    workflow_object.id_workflow
                )

            _workflow["workflow_class"] = workflow_object.workflow.name

        if isinstance(workflow_object.data, dict):
            record.update({"metadata": workflow_object.data})
        if isinstance(workflow_object.extra_data, dict):
            record.update({"_extra_data": workflow_object.extra_data})

        record["_workflow"] = _workflow
        return record

    def update_model(self):
        """Update model from current record."""
        if self.model is None:
            raise MissingModelError()

        self.workflow.data_type = self["_workflow"]["data_type"]
        self.workflow.status = ObjectStatus[self["_workflow"]["status"]]
        self.workflow.id_user = self["_workflow"]["id_user"]
        self.workflow.id_parent = self["_workflow"]["id_parent"]
        self.workflow.id_workflow = self["_workflow"]["id_workflow"]
        self.workflow.callback_pos = self["_workflow"]["workflow_position"]
        self.workflow.data = self['metadata']
        self.workflow.extra_data = self['_extra_data']
        self.workflow.save()

    def edit(self, *args, **kwargs):
        """Edit and save record (automatically indexed)."""
        record = request.json
        if record:
            self.update(record)
            self.commit()
        return self

    def restart(self, *args, **kwargs):
        """Resume execution from current task/callback in workflow."""
        if self.model is None:
            raise MissingModelError()

        if 'callback_pos' in kwargs:
            self.workflow.callback_pos = kwargs['callback_pos']
        else:
            self.workflow.callback_pos = 1

        self.workflow.save()
        db.session.commit()
        return resume.delay(
            oid=self.workflow.id,
            restart_point="restart_task"
        ).id

    def resume(self, *args, **kwargs):
        """Resume execution from next task/callback in workflow."""
        if self.model is None:
            raise MissingModelError()

        if 'callback_pos' in kwargs:
            self.workflow.callback_pos = kwargs['callback_pos']

        self.workflow.save()
        db.session.commit()
        return resume.delay(
            oid=self.id,
            restart_point="continue_next"
        ).id

    def resolve(self, *args, **kwargs):
        """Resolve an action if applicable."""
        if self.model is None:
            raise MissingModelError()

        action_name = self.workflow.get_action()
        if action_name:
            action_form = actions[action_name]
            return action_form.resolve(self.workflow, *args, **kwargs)

    @property
    def revision_id(self):
        """Special override as workflow object does not have revision."""
        return None

    @property
    def files(self):
        """Adapter for self.workflow files object."""
        return self.workflow.files



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""Serialization response factories.

Responsible for creating a HTTP response given the output of a serializer.
"""

from __future__ import absolute_import, print_function

from flask import current_app


def workflow_responsify(serializer, mimetype):
    """Create a Workflows-REST response serializer.

    :param serializer: Serializer function.
    :param mimetype: MIME type of response.
    """
    def view(workflow_object, code=200, headers=None):
        response = current_app.response_class(
            serializer.serialize(workflow_object),
            mimetype=mimetype)
        response.status_code = code

        if headers is not None:
            response.headers.extend(headers)
        return response
    return view


def action_responsify(serializer, mimetype):
    """Create a Workflows-REST action response serializer.

    :param serializer: Serializer function.
    :param mimetype: MIME type of response.
    """
    def view(data, code=200, headers=None):
        response = current_app.response_class(
            serializer.serialize_action(data),
            mimetype=mimetype)
        response.status_code = code

        if headers is not None:
            response.headers.extend(headers)
        return response
    return view


def file_responsify(serializer, mimetype):
    """Create a Workflows-REST file response serializer.

    :param serializer: Serializer function.
    :param mimetype: MIME type of response.
    """
    def view(data, code=200, headers=None):
        response = current_app.response_class(
            serializer.serialize_files(data),
            mimetype=mimetype)
        response.status_code = code

        if headers is not None:
            response.headers.extend(headers)
        return response
    return view


def search_responsify(serializer, mimetype):
    """Create a Workflows-REST search result response serializer.

    :param serializer: Serializer instance.
    :param mimetype: MIME type of response.
    """
    def view(search_result, code=200, headers=None, links=None):
        response = current_app.response_class(
            serializer.serialize_search(search_result,
                                        links=links),
            mimetype=mimetype)
        response.status_code = code
        if headers is not None:
            response.headers.extend(headers)
        return response
    return view



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Indexer overrides workflows UI interface."""

from __future__ import absolute_import, print_function

import pytz

from invenio_indexer.api import RecordIndexer


class WorkflowIndexer(RecordIndexer):
    """Special indexer for workflow objects."""

    @staticmethod
    def _prepare_record(record, index, doc_type):
        """Prepare the workflow object record for ES."""
        data = record.dumps()
        if record.model.created.tzinfo:
            data['_created'] = record.model.created.isoformat()
        else:
            data['_created'] = (
                pytz.utc.localize(record.model.created).isoformat()
                if record.model.created else None
            )
        if record.model.modified.tzinfo:
            data['_updated'] = record.model.modified.isoformat()
        else:
            data['_updated'] = (
                pytz.utc.localize(record.model.modified).isoformat()
                if record.model.modified else None
            )
        return data

    def index(self, record):
        """Index a record without version.

        NOTE: Can be removed when invenio-workflows model use versioning.

        :param record: Record instance.
        """
        index, doc_type = self.record_to_index(record)
        if not index or not doc_type:
            return
        return self.client.index(
            id=str(record.id),
            index=index,
            doc_type=doc_type,
            body=self._prepare_record(record, index, doc_type),
        )



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016 CERN.
#
# Invenio is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""Marshmallow based JSON serializer for records."""

from __future__ import absolute_import, print_function

from flask import current_app, json, request


class JSONSerializer(object):
    """JSON serializer for workflow object.

    Note: This serializer is not suitable for serializing large number of
    records.
    """

    @staticmethod
    def _format_args():
        """Get JSON dump indentation and separates."""
        if current_app.config['JSONIFY_PRETTYPRINT_REGULAR'] and \
                not request.is_xhr:
            return dict(
                indent=2,
                separators=(', ', ': '),
            )
        else:
            return dict(
                indent=None,
                separators=(',', ':'),
            )

    def serialize(self, workflow_ui_object):
        """Serialize a single workflow object.

        :param workflow_ui_object: workflow record instance.
        """
        return json.dumps(workflow_ui_object.dumps(), **self._format_args())

    def serialize_action(self, data):
        """Serialize an action response.

        :param workflow_ui_object: workflow record instance.
        """
        return json.dumps(dict(data), **self._format_args())

    def serialize_files(self, data):
        """Serialize files for a workflow object.

        :param data: some data object to serialize
        """
        from invenio_workflows.models import WorkflowFilesIterator
        if isinstance(data, WorkflowFilesIterator):
            return json.dumps([
                file_obj.dumps() for file_obj in data
            ], **self._format_args())
        else:
            return json.dumps(data.dumps(), **self._format_args())

    def serialize_search(self, search_result, links=None):
        """Serialize a search result.

        :param search_result: Elasticsearch search result.
        :param links: Dictionary of links to add to response.
        """
        return json.dumps(dict(
            hits=dict(
                hits=search_result['hits']['hits'],
                total=search_result['hits']['total'],
            ),
            links=links or {},
            aggregations=search_result.get('aggregations', dict()),
        ), **self._format_args())



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015, 2016 CERN.
#
# Invenio is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Invenio is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Invenio; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

"""Config variables for workflows UI module."""

from __future__ import absolute_import, print_function

from invenio_workflows_ui.search import terms_filter


WORKFLOWS_UI_URL = "/workflows"
WORKFLOWS_UI_API_URL = "/api/workflows/"
WORKFLOWS_UI_API_CLASS = "invenio_workflows_ui.api:WorkflowUIRecord"

WORKFLOWS_UI_TEMPLATE_CONTEXT_PROCESSORS = []
WORKFLOWS_UI_REST_ENDPOINT = dict(
    workflow_object_serializers={
        'application/json': ('invenio_workflows_ui.serializers'
                             ':json_serializer'),
    },
    search_serializers={
        'application/json': ('invenio_workflows_ui.serializers'
                             ':json_search_serializer'),
    },
    action_serializers={
        'application/json': ('invenio_workflows_ui.serializers'
                             ':json_action_serializer'),
    },
    bulk_action_serializers={
        'application/json': ('invenio_workflows_ui.serializers'
                             ':json_action_serializer'),
    },
    file_serializers={
        'application/json': ('invenio_workflows_ui.serializers'
                             ':json_file_serializer'),
    },
    list_route='/workflows/',
    item_route='/workflows/<object_id>',
    file_list_route='/workflows/<object_id>/files',
    file_item_route='/workflows/<object_id>/files/<path:key>',
    search_index="workflows",
    default_media_type='application/json',
    max_result_window=10000,
)

WORKFLOWS_UI_DATA_TYPES = dict(
    workflow=dict(
        search_index='workflows',
        search_type='record',
    ),
)

WORKFLOWS_UI_REST_FACETS = {
    "workflows": {
        "filters": {
            "status": terms_filter('_workflow.status'),
            "data_type": terms_filter('_workflow.data_type'),
            "workflow_name": terms_filter('_workflow.workflow_name'),
        },
        "aggs": {
            "status": {
                "terms": {
                    "field": "_workflow.status",
                    "size": 20
                }
            },
            "data_type": {
                "terms": {
                    "field": "_workflow.data_type",
                    "size": 20
                }
            },
            "workflow_name": {
                "terms": {
                    "field": "_workflow.workflow_name",
                    "size": 20
                }
            },
        }
    }
}

WORKFLOWS_UI_REST_SORT_OPTIONS = {
    "workflows": {
        "bestmatch": {
            "title": 'Best match',
            "fields": ['_score'],
            "default_order": 'desc',
            "order": 1,
        },
        "mostrecent": {
            "title": 'Most recent',
            "fields": ['_workflow.modified'],
            "default_order": 'desc',
            "order": 2,
        },
    },
}

WORKFLOWS_UI_REST_DEFAULT_SORT = {
    "workflows": {
        "query": "-bestmatch",
        "noquery": "-mostrecent"
    }
}


WORKFLOWS_UI_CACHE_PREFIX = "WorkflowsUI::"
WORKFLOWS_UI_LIST_TEMPLATE = "invenio_workflows_ui/list.html"
WORKFLOWS_UI_DETAILS_TEMPLATE = "invenio_workflows_ui/details.html"
WORKFLOWS_UI_INDEX_TEMPLATE = "invenio_workflows_ui/index.html"

WORKFLOWS_UI_JSTEMPLATE_RESULTS = (
    "templates/invenio_workflows_ui/results.html"
)
WORKFLOWS_UI_JSTEMPLATE_COUNT = (
    "templates/invenio_search_ui/count.html"
)
WORKFLOWS_UI_JSTEMPLATE_PAGINATION = (
    "templates/invenio_search_ui/pagination.html"
)
WORKFLOWS_UI_JSTEMPLATE_SELECT_BOX = (
    "templates/invenio_search_ui/selectbox.html"
)
WORKFLOWS_UI_JSTEMPLATE_SORT_ORDER = (
    "templates/invenio_search_ui/togglebutton.html"
)
WORKFLOWS_UI_JSTEMPLATE_ERROR = "templates/invenio_search_ui/error.html"
WORKFLOWS_UI_JSTEMPLATE_LOADING = "templates/invenio_search_ui/loading.html"
WORKFLOWS_UI_JSTEMPLATE_FACETS = (
    "node_modules/invenio-search-js/dist/templates/facets.html"
)
